@@ -1,6 +1,7 @@
 ##############################################################################
 #
-# Copyright (c) 2010 Vifib SARL and Contributors. All Rights Reserved.
+# Copyright (c) 2010, 2011, 2012 Vifib SARL and Contributors.
+# All Rights Reserved.
 #
 # WARNING: This program as such is intended to be used by professional
 # programmers who take the whole responsibility of assessing all potential
@@ -10,8 +11,8 @@
 # Service Company
 #
 # This program is Free Software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 3
+# modify it under the terms of the GNU Lesser General Public License
+# as published by the Free Software Foundation; either version 2.1
 # of the License, or (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -19,7 +20,7 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
-# You should have received a copy of the GNU General Public License
+# You should have received a copy of the GNU Lesser General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #
@@ -34,12 +35,24 @@ class IException(Interface):
   Classes which implement IException are used to report errors.
   """
 
+class IServerError(IException):
+  """
+  Classes which implement IServerError are used to report unexpected error
+  from the slap server.
+  """
+
 class INotFoundError(IException):
   """
   Classes which implement INotFoundError are used to report missing
   informations on the slap server.
   """
 
+class IResourceNotReady(IException):
+  """
+  Classes which implement IResourceNotReady are used to report resource not
+  ready on the slap server.
+  """
+
 class IUnauthorized(IException):
   """
   Classes which implement IUnauthorized are used to report missing
@@ -52,10 +65,10 @@ class IRequester(Interface):
   slapgrid server.
   """
 
-  def request(software_release, software_type, partition_reference, 
+  def request(software_release, software_type, partition_reference,
               shared=False, partition_parameter_kw=None, filter_kw=None):
     """
-    Request software release instanciation to slapgrid server.
+    Request software release instantiation to slapgrid server.
 
     Returns a new computer partition document, where this sofware release will
     be installed.
@@ -92,13 +105,13 @@ class IBuildoutController(Interface):
 
   def available():
     """
-    Notify (to the slapgrid server) that the software instance is 
+    Notify (to the slapgrid server) that the software instance is
     available.
     """
 
   def building():
     """
-    Notify (to the slapgrid server) that the buildout is not 
+    Notify (to the slapgrid server) that the buildout is not
     available and under creation.
     """
 
@@ -121,6 +134,26 @@ class ISoftwareRelease(IBuildoutController):
     Returns a string representing the uri of the software release.
     """
 
+  def getComputerId():
+    """
+    Returns a string representing the identifier of the computer where the SR
+    is installed.
+    """
+
+  def getState():
+    """
+    Returns a string representing the expected state of the software
+    installation.
+
+    The result can be: available, destroyed
+    """
+
+  def destroyed():
+    """
+    Notify (to the slapgrid server) that the software installation has
+    been correctly destroyed.
+    """
+
 class IComputerPartition(IBuildoutController, IRequester):
   """
   Computer Partition interface specification
@@ -132,13 +165,13 @@ class IComputerPartition(IBuildoutController, IRequester):
 
   def stopped():
     """
-    Notify (to the slapgrid server) that the software instance is 
+    Notify (to the slapgrid server) that the software instance is
     available and stopped.
     """
 
   def started():
     """
-    Notify (to the slapgrid server) that the software instance is 
+    Notify (to the slapgrid server) that the software instance is
     available and started.
     """
 
@@ -176,6 +209,13 @@ class IComputerPartition(IBuildoutController, IRequester):
     profile.
     """
 
+  def getConnectionParameterDict():
+    """
+    Returns a dictionary of connection parameters.
+
+    The contained values are connection parameters of a compute partition.
+    """
+
   def setUsage(usage_log):
     """
     Associate a usage log to the computer partition.
@@ -194,6 +234,65 @@ class IComputerPartition(IBuildoutController, IRequester):
     log -- a text explaining why the method was called
     """
 
+  def getCertificate():
+    """
+    Returns a dictionnary containing the authentification certificates
+    associated to the computer partition.
+    The dictionnary keys are:
+      key -- value is a SSL key
+      certificate -- value is a SSL certificate
+
+    Raise an INotFoundError if no software release is associated.
+    """
+
+  def setConnectionDict(connection_dict, slave_reference=None):
+    """
+    Store the connection parameters associated to a partition.
+
+    connection_dict -- dictionary of parameter used to fill the
+                              connection dict of the partition.
+
+    slave_reference -- current reference of the slave instance to modify
+    """
+
+  def getInstanceParameter(key):
+    """
+    Returns the instance parameter associated to the key.
+
+    Raise an INotFoundError if no key is defined.
+
+    key -- a string name of the parameter
+    """
+
+  def getConnectionParameter(key):
+    """
+    Return the connection parameter associate to the key.
+
+    Raise an INotFoundError if no key is defined.
+
+    key -- a string name of the parameter
+    """
+
+  def rename(partition_reference, slave_reference=None):
+    """
+    Change the partition reference of a partition
+
+    partition_reference -- new local reference of the instance used by the recipe
+                           to identify the instances.
+
+    slave_reference -- current reference of the slave instance to modify
+    """
+
+  def getStatus():
+    """
+    Returns a dictionnary containing the latest status of the 
+    computer partition.
+    The dictionnary keys are:
+      user -- user who reported the latest status
+      created_at -- date of the status
+      text -- message log of the status
+    """
+
 class IComputer(Interface):
   """
   Computer interface specification
@@ -203,7 +302,7 @@ class IComputer(Interface):
   installed.
   """
 
-  def getSoftwareReleaseList(): 
+  def getSoftwareReleaseList():
     """
     Returns the list of software release which has to be supplied by the
     computer.
@@ -211,7 +310,7 @@ class IComputer(Interface):
     Raise an INotFoundError if computer_guid doesn't exist.
     """
 
-  def getComputerPartitionList(): 
+  def getComputerPartitionList():
     """
     Returns the list of configured computer partitions associated to this
     computer.
@@ -221,7 +320,7 @@ class IComputer(Interface):
 
   def reportUsage(computer_partition_list):
     """
-    Report the computer usage to the slapgrid server. 
+    Report the computer usage to the slapgrid server.
     IComputerPartition.setUsage has to be called on each computer partition to
     define each usage.
 
@@ -238,6 +337,41 @@ class IComputer(Interface):
     log -- a text explaining why the method was called
     """
 
+  def updateConfiguration(configuration_xml):
+    """
+    Report the current computer configuration.
+
+    configuration_xml -- computer XML description generated by slapformat
+    """
+
+  def getStatus():
+    """
+    Returns a dictionnary containing the latest status of the 
+    computer.
+    The dictionnary keys are:
+      user -- user who reported the latest status
+      created_at -- date of the status
+      text -- message log of the status
+    """
+
+  def generateCertificate():
+    """
+    Returns a dictionnary containing the new certificate files for 
+    the computer.
+    The dictionnary keys are:
+      key -- key file
+      certificate -- certificate file
+
+    Raise ValueError is another certificate is already valid.
+    """
+
+  def revokeCertificate():
+    """
+    Revoke current computer certificate.
+
+    Raise ValueError is there is not valid certificate.
+    """
+
 class IOpenOrder(IRequester):
   """
   Open Order interface specification
@@ -246,6 +380,15 @@ class IOpenOrder(IRequester):
   is requested by a given client.
   """
 
+  def requestComputer(computer_reference):
+    """
+    Request a computer to slapgrid server.
+
+    Returns a new computer document.
+
+    computer_reference -- local reference of the computer
+    """
+
 class ISupply(Interface):
   """
   Supply interface specification
@@ -261,7 +404,7 @@ class ISupply(Interface):
     software_release -- uri of the software release
                         which has to be instanciated
 
-    computer_guid -- the identifier of the computer inside the slapgrid 
+    computer_guid -- the identifier of the computer inside the slapgrid
                      server.
     """
 
@@ -269,7 +412,7 @@ class slap(Interface):
   """
   Initialise slap connection to the slapgrid server
 
-  Slapgrid server URL is defined during the slap library installation, 
+  Slapgrid server URL is defined during the slap library installation,
   as recipes should not use another server.
   """
 
@@ -297,7 +440,7 @@ class slap(Interface):
 
     computer_guid -- the identifier of the computer inside the slapgrid server.
 
-    partition_id -- the identifier of the computer partition inside the 
+    partition_id -- the identifier of the computer partition inside the
                     slapgrid server.
 
     Raise an INotFoundError if computer_guid doesn't exist.
@@ -314,8 +457,8 @@ class slap(Interface):
     """
     Instanciate an open order in the slap library.
     """
-    
+
   def registerSupply():
-    """   
+    """
     Instanciate a supply in the slap library.
     """
