@@ -1,7 +1,8 @@
 # -*- coding: utf-8 -*-
 ##############################################################################
 #
-# Copyright (c) 2010 Vifib SARL and Contributors. All Rights Reserved.
+# Copyright (c) 2010, 2011, 2012 Vifib SARL and Contributors.
+# All Rights Reserved.
 #
 # WARNING: This program as such is intended to be used by professional
 # programmers who take the whole responsibility of assessing all potential
@@ -11,8 +12,8 @@
 # Service Company
 #
 # This program is Free Software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 3
+# modify it under the terms of the GNU Lesser General Public License
+# as published by the Free Software Foundation; either version 2.1
 # of the License, or (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -20,14 +21,14 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
-# You should have received a copy of the GNU General Public License
+# You should have received a copy of the GNU Lesser General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #
 ##############################################################################
 __all__ = ["slap", "ComputerPartition", "Computer", "SoftwareRelease",
            "Supply", "OpenOrder", "NotFoundError", "Unauthorized",
-           "ResourceNotReady"]
+           "ResourceNotReady", "ServerError"]
 
 from interface import slap as interface
 from xml_marshaller import xml_marshaller
@@ -42,6 +43,8 @@ import zope.interface
 Simple, easy to (un)marshall classes for slap client/server communication
 """
 
+DEFAULT_SOFTWARE_TYPE = 'default'
+
 # httplib.HTTPSConnection with key verification
 class HTTPSConnectionCA(httplib.HTTPSConnection):
   """Patched version of HTTPSConnection which verifies server certificate"""
@@ -64,7 +67,11 @@ class HTTPSConnectionCA(httplib.HTTPSConnection):
 
 
 class SlapDocument:
-  pass
+  def __init__(self, connection_helper=None):
+    if connection_helper is not None:
+      # Do not require connection_helper to be provided, but when it's not,
+      # cause failures when accessing _connection_helper property.
+      self._connection_helper = connection_helper
 
 class SoftwareRelease(SlapDocument):
   """
@@ -79,6 +86,7 @@ class SoftwareRelease(SlapDocument):
 
     XXX **kw args only kept for compatibility
     """
+    SlapDocument.__init__(self, kw.pop('connection_helper', None))
     self._software_instance_list = []
     if software_release is not None:
       software_release = software_release.encode('UTF-8')
@@ -88,30 +96,39 @@ class SoftwareRelease(SlapDocument):
   def __getinitargs__(self):
     return (self._software_release, self._computer_guid, )
 
+  def getComputerId(self):
+    if not self._computer_guid:
+      raise NameError('computer_guid has not been defined.')
+    else:
+      return self._computer_guid
+
+  def getURI(self):
+    if not self._software_release:
+      raise NameError('software_release has not been defined.')
+    else:
+      return self._software_release
+
   def error(self, error_log):
     # Does not follow interface
     self._connection_helper.POST('/softwareReleaseError', {
-      'url': self._software_release,
-      'computer_id' : self._computer_guid,
+      'url': self.getURI(),
+      'computer_id' : self.getComputerId(),
       'error_log': error_log})
 
-  def getURI(self):
-    return self._software_release
-
   def available(self):
     self._connection_helper.POST('/availableSoftwareRelease', {
-      'url': self._software_release, 
-      'computer_id': self._computer_guid})
+      'url': self.getURI(),
+      'computer_id': self.getComputerId()})
 
   def building(self):
     self._connection_helper.POST('/buildingSoftwareRelease', {
-      'url': self._software_release, 
-      'computer_id': self._computer_guid})
+      'url': self.getURI(),
+      'computer_id': self.getComputerId()})
 
   def destroyed(self):
     self._connection_helper.POST('/destroyedSoftwareRelease', {
-      'url': self._software_release, 
-      'computer_id': self._computer_guid})
+      'url': self.getURI(),
+      'computer_id': self.getComputerId()})
 
   def getState(self):
     return getattr(self, '_requested_state', 'available')
@@ -132,10 +149,10 @@ class SoftwareInstance(SlapDocument):
 """Exposed exceptions"""
 # XXX Why do we need to expose exceptions?
 class ResourceNotReady(Exception):
-  pass
+  zope.interface.implements(interface.IResourceNotReady)
 
 class ServerError(Exception):
-  pass
+  zope.interface.implements(interface.IServerError)
 
 class NotFoundError(Exception):
   zope.interface.implements(interface.INotFoundError)
@@ -148,12 +165,14 @@ class Supply(SlapDocument):
   zope.interface.implements(interface.ISupply)
 
   def supply(self, software_release, computer_guid=None, state='available'):
-    # Note: Zope accepts additional arguments, so state in case of older
-    #       servers will be just ignored.
-    self._connection_helper.POST('/supplySupply', {
-      'url': software_release,
-      'computer_id': computer_guid,
-      'state': state})
+    try:
+      self._connection_helper.POST('/supplySupply', {
+        'url': software_release,
+        'computer_id': computer_guid,
+        'state': state})
+    except NotFoundError:
+      raise NotFoundError("Computer %s has not been found by SlapOS Master."
+          % computer_guid)
 
 class OpenOrder(SlapDocument):
 
@@ -171,47 +190,74 @@ class OpenOrder(SlapDocument):
         'partition_reference': partition_reference,
         'partition_parameter_xml': xml_marshaller.dumps(partition_parameter_kw),
         'filter_xml': xml_marshaller.dumps(filter_kw),
+        # XXX Cedric: Why state and shared are marshalled? First is a string
+        #             And second is a boolean.
         'state': xml_marshaller.dumps(state),
         'shared_xml': xml_marshaller.dumps(shared),
       }
     if software_type is not None:
       request_dict['software_type'] = software_type
+    else:
+      # Let's enforce a default software type
+      request_dict['software_type'] = DEFAULT_SOFTWARE_TYPE
     try:
       self._connection_helper.POST('/requestComputerPartition', request_dict)
     except ResourceNotReady:
-      return ComputerPartition(request_dict=request_dict)
+      return ComputerPartition(
+        request_dict=request_dict,
+        connection_helper=self._connection_helper,
+      )
     else:
       xml = self._connection_helper.response.read()
       software_instance = xml_marshaller.loads(xml)
       computer_partition = ComputerPartition(
         software_instance.slap_computer_id.encode('UTF-8'),
-        software_instance.slap_computer_partition_id.encode('UTF-8'))
+        software_instance.slap_computer_partition_id.encode('UTF-8'),
+        connection_helper=self._connection_helper,
+      )
       if shared:
         computer_partition._synced = True
         computer_partition._connection_dict = software_instance._connection_dict
         computer_partition._parameter_dict = software_instance._parameter_dict
       return computer_partition
 
+  def requestComputer(self, computer_reference):
+    """
+    Requests a computer.
+    """
+    self._connection_helper.POST('/requestComputer',
+      {'computer_title': computer_reference})
+    xml = self._connection_helper.response.read()
+    computer = xml_marshaller.loads(xml)
+    computer._connection_helper = self._connection_helper
+    return computer
+
 def _syncComputerInformation(func):
   """
   Synchronize computer object with server information
   """
   def decorated(self, *args, **kw):
-    computer = self._connection_helper.getComputerInformation(self._computer_id)
+    if getattr(self, '_synced', 0):
+      return func(self, *args, **kw)
+    computer = self._connection_helper.getFullComputerInformation(self._computer_id)
     for key, value in computer.__dict__.items():
       if isinstance(value, unicode):
         # convert unicode to utf-8
         setattr(self, key, value.encode('utf-8'))
       else:
         setattr(self, key, value)
+    setattr(self, '_synced', True)
+    for computer_partition in self.getComputerPartitionList():
+      setattr(computer_partition, '_synced', True)
     return func(self, *args, **kw)
-  return decorated 
+  return decorated
 
 class Computer(SlapDocument):
 
   zope.interface.implements(interface.IComputer)
 
-  def __init__(self, computer_id):
+  def __init__(self, computer_id, connection_helper=None):
+    SlapDocument.__init__(self, connection_helper)
     self._computer_id = computer_id
 
   def __getinitargs__(self):
@@ -225,11 +271,15 @@ class Computer(SlapDocument):
 
     Raise an INotFoundError if computer_guid doesn't exist.
     """
+    for software_relase in self._software_release_list:
+      software_relase._connection_helper = self._connection_helper
     return self._software_release_list
 
   @_syncComputerInformation
   def getComputerPartitionList(self):
-    return [x for x in self._computer_partition_list if x._need_modification]
+    for computer_partition in self._computer_partition_list:
+      computer_partition._connection_helper = self._connection_helper
+    return [x for x in self._computer_partition_list ]
 
   def reportUsage(self, computer_usage):
     if computer_usage == "":
@@ -248,6 +298,21 @@ class Computer(SlapDocument):
       'computer_id': self._computer_id,
       'message': message})
 
+  def getStatus(self):
+    self._connection_helper.GET(
+        '/getComputerStatus?computer_id=%s' % self._computer_id)
+    return xml_marshaller.loads(self._connection_helper.response.read())
+
+  def revokeCertificate(self):
+    self._connection_helper.POST('/revokeComputerCertificate', {
+      'computer_id': self._computer_id})
+
+  def generateCertificate(self):
+    self._connection_helper.POST('/generateComputerCertificate', {
+      'computer_id': self._computer_id})
+    xml = self._connection_helper.response.read()
+    return xml_marshaller.loads(xml)
+
 def _syncComputerPartitionInformation(func):
   """
   Synchronize computer partition object with server information
@@ -255,14 +320,19 @@ def _syncComputerPartitionInformation(func):
   def decorated(self, *args, **kw):
     if getattr(self, '_synced', 0):
       return func(self, *args, **kw)
-    computer = self._connection_helper.getComputerInformation(self._computer_id)
+    if not self._computer_id:
+      # XXX Is it only in case of requesting instance?
+      raise ResourceNotReady("Instance is not ready yet.")
+    # XXX: This is a ugly way to keep backward compatibility,
+    # We should stablise slap library soon.
+    computer = self._connection_helper.getFullComputerInformation(self._computer_id)
     found_computer_partition = None
     for computer_partition in computer._computer_partition_list:
       if computer_partition.getId() == self.getId():
         found_computer_partition = computer_partition
         break
     if found_computer_partition is None:
-      raise NotFoundError("No software release information for partition %s" %
+      raise NotFoundError("No information for partition %s" %
           self.getId())
     else:
       for key, value in found_computer_partition.__dict__.items():
@@ -290,6 +360,7 @@ def _syncComputerPartitionInformation(func):
           setattr(self, key, new_dict)
         else:
           setattr(self, key, value)
+    setattr(self, '_synced', True)
     return func(self, *args, **kw)
   return decorated
 
@@ -298,7 +369,9 @@ class ComputerPartition(SlapDocument):
 
   zope.interface.implements(interface.IComputerPartition)
 
-  def __init__(self, computer_id=None, partition_id=None, request_dict=None):
+  def __init__(self, computer_id=None, partition_id=None, request_dict=None,
+      connection_helper=None):
+    SlapDocument.__init__(self, connection_helper)
     if request_dict is not None and (computer_id is not None or
         partition_id is not None):
       raise TypeError('request_dict conflicts with computer_id and '
@@ -313,11 +386,6 @@ class ComputerPartition(SlapDocument):
   def __getinitargs__(self):
     return (self._computer_id, self._partition_id, )
 
-  # XXX: As request is decorated with _syncComputerPartitionInformation it
-  #      will raise ResourceNotReady really early -- just after requesting,
-  #      and not when try to access to real partition is required.
-  #      To have later raising (like in case of calling methods), the way how
-  #      Computer Partition data are fetch from server shall be delayed
   @_syncComputerPartitionInformation
   def request(self, software_release, software_type, partition_reference,
               shared=False, partition_parameter_kw=None, filter_kw=None,
@@ -334,6 +402,10 @@ class ComputerPartition(SlapDocument):
       raise ValueError("Unexpected type of filter_kw '%s'" % \
                        filter_kw)
 
+    # Let enforce a default software type
+    if software_type is None:
+      software_type = DEFAULT_SOFTWARE_TYPE
+
     request_dict = { 'computer_id': self._computer_id,
         'computer_partition_id': self._partition_id,
         'software_release': software_release,
@@ -348,13 +420,18 @@ class ComputerPartition(SlapDocument):
     try:
       self._connection_helper.POST('/requestComputerPartition', request_dict)
     except ResourceNotReady:
-      return ComputerPartition(request_dict=request_dict)
+      return ComputerPartition(
+        request_dict=request_dict,
+        connection_helper=self._connection_helper,
+      )
     else:
       xml = self._connection_helper.response.read()
       software_instance = xml_marshaller.loads(xml)
       computer_partition = ComputerPartition(
         software_instance.slap_computer_id.encode('UTF-8'),
-        software_instance.slap_computer_partition_id.encode('UTF-8'))
+        software_instance.slap_computer_partition_id.encode('UTF-8'),
+        connection_helper=self._connection_helper,
+      )
       if shared:
         computer_partition._synced = True
         computer_partition._connection_dict = getattr(software_instance,
@@ -366,47 +443,47 @@ class ComputerPartition(SlapDocument):
   def building(self):
     self._connection_helper.POST('/buildingComputerPartition', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id})
+      'computer_partition_id': self.getId()})
 
   def available(self):
     self._connection_helper.POST('/availableComputerPartition', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id})
+      'computer_partition_id': self.getId()})
 
   def destroyed(self):
     self._connection_helper.POST('/destroyedComputerPartition', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
+      'computer_partition_id': self.getId(),
       })
 
   def started(self):
     self._connection_helper.POST('/startedComputerPartition', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
+      'computer_partition_id': self.getId(),
       })
 
   def stopped(self):
     self._connection_helper.POST('/stoppedComputerPartition', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
+      'computer_partition_id': self.getId(),
       })
 
   def error(self, error_log):
     self._connection_helper.POST('/softwareInstanceError', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
+      'computer_partition_id': self.getId(),
       'error_log': error_log})
 
   def bang(self, message):
     self._connection_helper.POST('/softwareInstanceBang', {
       'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
+      'computer_partition_id': self.getId(),
       'message': message})
 
   def rename(self, new_name, slave_reference=None):
     post_dict = dict(
       computer_id=self._computer_id,
-      computer_partition_id=self._partition_id,
+      computer_partition_id=self.getId(),
       new_name=new_name,
     )
     if slave_reference is not None:
@@ -414,6 +491,8 @@ class ComputerPartition(SlapDocument):
     self._connection_helper.POST('/softwareInstanceRename', post_dict)
 
   def getId(self):
+    if not self._partition_id:
+      raise ResourceNotReady()
     return self._partition_id
 
   @_syncComputerPartitionInformation
@@ -424,23 +503,36 @@ class ComputerPartition(SlapDocument):
   def getInstanceParameterDict(self):
     return getattr(self, '_parameter_dict', None) or {}
 
+  @_syncComputerPartitionInformation
+  def getConnectionParameterDict(self):
+    return getattr(self, '_connection_dict', None) or {}
+
   @_syncComputerPartitionInformation
   def getSoftwareRelease(self):
     """
     Returns the software release associate to the computer partition.
     """
-    if self._software_release_document is None:
+    if not getattr(self, '_software_release_document', None):
       raise NotFoundError("No software release information for partition %s" %
           self.getId())
     else:
       return self._software_release_document
 
   def setConnectionDict(self, connection_dict, slave_reference=None):
-    self._connection_helper.POST('/setComputerPartitionConnectionXml', {
-      'computer_id': self._computer_id,
-      'computer_partition_id': self._partition_id,
-      'connection_xml': xml_marshaller.dumps(connection_dict),
-      'slave_reference': slave_reference})
+    if self.getConnectionParameterDict() != connection_dict:
+      self._connection_helper.POST('/setComputerPartitionConnectionXml', {
+          'computer_id': self._computer_id,
+          'computer_partition_id': self._partition_id,
+          'connection_xml': xml_marshaller.dumps(connection_dict),
+          'slave_reference': slave_reference})
+
+  @_syncComputerPartitionInformation
+  def getInstanceParameter(self, key):
+    parameter_dict = getattr(self, '_parameter_dict', None) or {}
+    if key in parameter_dict:
+      return parameter_dict[key]
+    else:
+      raise NotFoundError("%s not found" % key)
 
   @_syncComputerPartitionInformation
   def getConnectionParameter(self, key):
@@ -460,20 +552,11 @@ class ComputerPartition(SlapDocument):
         'computer_partition_id=%s' % (self._computer_id, self._partition_id))
     return xml_marshaller.loads(self._connection_helper.response.read())
 
-# def lazyMethod(func):
-#   """
-#   Return a function which stores a computed value in an instance
-#   at the first call.
-#   """
-#   key = '_cache_' + str(id(func))
-#   def decorated(self, *args, **kw):
-#     try:
-#       return getattr(self, key)
-#     except AttributeError:
-#       result = func(self, *args, **kw)
-#       setattr(self, key, result)
-#       return result
-#   return decorated 
+  def getStatus(self):
+    self._connection_helper.GET(
+        '/getComputerPartitionStatus?computer_id=%s&'
+        'computer_partition_id=%s' % (self._computer_id, self._partition_id))
+    return xml_marshaller.loads(self._connection_helper.response.read())
 
 class ConnectionHelper:
   error_message_timeout = "\nThe connection timed out. Please try again later."
@@ -497,6 +580,24 @@ class ConnectionHelper:
     self.GET('/getComputerInformation?computer_id=%s' % computer_id)
     return xml_marshaller.loads(self.response.read())
 
+  def getFullComputerInformation(self, computer_id):
+    """
+    Retrieve from SlapOS Master Computer instance containing all needed
+    informations (Software Releases, Computer Partitions, ...).
+    """
+    method = '/getFullComputerInformation?computer_id=%s' % computer_id
+    if not computer_id:
+      # XXX-Cedric: should raise something smarter than "NotFound".
+      raise NotFoundError(method)
+    try:
+      self.GET(method)
+    except NotFoundError:
+      # XXX: This is a ugly way to keep backward compatibility,
+      # We should stablise slap library soon.
+      self.GET('/getComputerInformation?computer_id=%s' % computer_id)
+
+    return xml_marshaller.loads(self.response.read())
+
   def connect(self):
     connection_dict = dict(
         host=self.host)
@@ -522,6 +623,8 @@ class ConnectionHelper:
           raise socket.error(str(e) + self.error_message_timeout)
         raise ssl.SSLError(str(e) + self.ssl_error_message_connect_fail)
       except socket.error, e:
+        if e.message == "timed out":
+          raise socket.error(str(e) + self.error_message_timeout)
         raise socket.error(self.error_message_connect_fail + str(e))
       # check self.response.status and raise exception early
       if self.response.status == httplib.REQUEST_TIMEOUT:
@@ -575,69 +678,62 @@ class slap:
 
   def initializeConnection(self, slapgrid_uri, key_file=None, cert_file=None,
       master_ca_file=None, timeout=60):
-    self._initialiseConnectionHelper(slapgrid_uri, key_file, cert_file,
-        master_ca_file, timeout)
-
-  def _initialiseConnectionHelper(self, slapgrid_uri, key_file, cert_file,
-      master_ca_file, timeout):
-    SlapDocument._slapgrid_uri = slapgrid_uri
     scheme, netloc, path, query, fragment = urlparse.urlsplit(
-        SlapDocument._slapgrid_uri)
+        slapgrid_uri)
     if not(query == '' and fragment == ''):
       raise AttributeError('Passed URL %r issue: not parseable'%
-          SlapDocument._slapgrid_uri)
-    if scheme not in ('http', 'https'):
-      raise AttributeError('Passed URL %r issue: there is no support for %r p'
-          'rotocol' % (SlapDocument._slapgrid_uri, scheme))
+          slapgrid_uri)
 
     if scheme == 'http':
       connection_wrapper = httplib.HTTPConnection
-    else:
+    elif scheme == 'https':
       if master_ca_file is not None:
         connection_wrapper = HTTPSConnectionCA
       else:
         connection_wrapper = httplib.HTTPSConnection
-    slap._connection_helper = \
-      SlapDocument._connection_helper = ConnectionHelper(connection_wrapper,
+    else:
+      raise AttributeError('Passed URL %r issue: there is no support for %r p'
+          'rotocol' % (slapgrid_uri, scheme))
+    self._connection_helper = ConnectionHelper(connection_wrapper,
           netloc, path, key_file, cert_file, master_ca_file, timeout)
 
-  def _register(self, klass, *registration_argument_list):
-    if len(registration_argument_list) == 1 and type(
-        registration_argument_list[0]) == type([]):
-      # in case if list is explicitly passed and there is only one
-      # argument in registration convert it to list
-      registration_argument_list = registration_argument_list[0]
-    document = klass(*registration_argument_list)
-    return document
-
+  # XXX-Cedric: this method is never used and thus should be removed.
   def registerSoftwareRelease(self, software_release):
     """
     Registers connected representation of software release and
     returns SoftwareRelease class object
     """
-    return SoftwareRelease(software_release=software_release)
+    return SoftwareRelease(software_release=software_release,
+      connection_helper=self._connection_helper,
+    )
 
   def registerComputer(self, computer_guid):
     """
     Registers connected representation of computer and
     returns Computer class object
     """
-    self.computer_guid = computer_guid
-    return self._register(Computer, computer_guid)
+    return Computer(computer_guid, connection_helper=self._connection_helper)
 
   def registerComputerPartition(self, computer_guid, partition_id):
     """
     Registers connected representation of computer partition and
     returns Computer Partition class object
     """
+    if not computer_guid or not partition_id:
+      # XXX-Cedric: should raise something smarter than NotFound
+      raise NotFoundError
+
     self._connection_helper.GET('/registerComputerPartition?' \
         'computer_reference=%s&computer_partition_reference=%s' % (
           computer_guid, partition_id))
-    xml = self._connection_helper.response.read()
-    return xml_marshaller.loads(xml)
+    result = xml_marshaller.loads(self._connection_helper.response.read())
+    # XXX: dirty hack to make computer partition usable. xml_marshaller is too
+    # low-level for our needs here.
+    result._connection_helper = self._connection_helper
+    return result
 
   def registerOpenOrder(self):
-    return OpenOrder()
+    return OpenOrder(connection_helper=self._connection_helper)
 
   def registerSupply(self):
-    return Supply()
+    return Supply(connection_helper=self._connection_helper)
